/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.engine.source;

import java.io.ByteArrayInputStream;
import java.io.Serializable;
import java.security.Principal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.exchange.xml.XmlParser;

import sync4j.foundation.pdi.converter.*;
import sync4j.foundation.pdi.parser.*;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.tools.Base64;

import sync4j.exchange.items.calendar.model.Calendar;
import sync4j.exchange.items.calendar.manager.CalendarManager;
import sync4j.exchange.items.calendar.CalendarParseException;
import sync4j.exchange.DataAccessException;

/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server calendar items
 *
 * version $Id: ExchangeCalendarSyncSource.java,v 1.23 2005/09/29 18:16:05 nichele Exp $
 */
public class ExchangeCalendarSyncSource
extends ExchangeSyncSource
implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // --------------------------------------------------------------- Private data

    private CalendarManager cm      = null  ;

    // --------------------------------------------------------------- Constructors

    public ExchangeCalendarSyncSource() {
    }

    // --------------------------------------------------------------- Public Methods

   /**
    * @see SyncSource
    */
    public SyncItem setSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        Calendar calendar       = null  ;
        Calendar newCalendar    = null  ;

        String   href           = null  ;
        String   content        = null  ;
        String   itemKey        = null  ;
        String   username       = null  ;
        String   credentials    = null  ;

        boolean  isAddCalendar  = false ;

        try {

            username    = ((Sync4jPrincipal)principal).getUsername()           ;

            credentials = ((Sync4jPrincipal)principal).getEncodedCredentials() ;

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // search href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey);

            if (href == null) {
                //
                // if href not found, new item is created
                // default href = this.CurrentTimesMillis()
                //
                href = String.valueOf(System.currentTimeMillis()) +
                       EXCHANGE_HREF_EXTENSION;
                isAddCalendar = true;
            }

            if (log.isLoggable(Level.FINEST)) {
                if (!isAddCalendar) {
                    log.finest(itemKey + " already exists: updating");
                } else {
                    log.finest(itemKey + " is new: adding");
                }
            }

            calendar = getCalendarFromSyncItem(syncItem);

            calendar.setHref(href);

            cm = getCalendarManager();

            newCalendar = cm.setCalendar (calendar    ,
                                          username    ,
                                          credentials ,
                                          exchangeFolder);

            if (isAddCalendar) {
                toChangedItems(newCalendar.getId(), ITEM_ADD);
            } else {
                toChangedItems(newCalendar.getId(), ITEM_UPDATE);
            }


            syncItem = getSyncItem(newCalendar);

            return syncItem;

        } catch (Exception e) {
            if (newCalendar != null) {
                if (isAddCalendar) {
                    toChangedItems(newCalendar.getId(), ITEM_ADD_ERROR);
                } else {
                    toChangedItems(newCalendar.getId(), ITEM_UPDATE_ERROR);
                }
            }
            throw new SyncSourceException( "Error setting the item " +
                                           syncItem + " " + e.getMessage(), e) ;
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList() ;
        ArrayList ret              = new ArrayList() ;

        for (int i=0, l = syncItems.length; i < l; ++i) {
            try {
                ret.add(setSyncItem(principal, syncItems[i]));
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException("Error setting the following items: " +
                                          syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource " +
                       sourceURI              +
                       " - set syncItems");
        }

        return (SyncItem[])ret.toArray(new SyncItem[] {});

    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromId(Principal principal,
                                      SyncItemKey syncItemKey)
    throws SyncSourceException{

        SyncItem[] syncItems =
            getSyncItemsFromIds(principal, new SyncItemKey[] {syncItemKey});

        if ((syncItems == null) || (syncItems.length == 0)) {
            return null;
        }

        return syncItems[0];

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromIds(Principal principal,
                                          SyncItemKey[] syncItemKeys)
    throws SyncSourceException {

        ArrayList  syncItems    = null ;
        Calendar   calendar     = null ;

        String     username     = null ;
        String     credentials  = null ;

        String     id           = null ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        syncItems = new ArrayList();

        int l = syncItemKeys.length;

        for (int i=0; ((syncItemKeys != null) && (i < l)); ++i) {

            id = syncItemKeys[i].getKeyAsString();

            try {
                cm = getCalendarManager();
                calendar = cm.getCalendarById(username    ,
                                              credentials ,
                                              id          ,
                                              this.exchangeFolder);
            } catch (DataAccessException e) {

                Throwable previous = e.getCause();

                if (previous instanceof NotFoundException) {

                    Logger log = Sync4jLogger.getLogger("source");
                    if (log.isLoggable(Level.SEVERE)) {
                        log.severe
                            ("Calendar not found while reading Exchange database: " +
                              e.getMessage()                                        );
                    }

                    Logger.
                        getLogger(LOG_NAME).
                            throwing(getClass().
                                getName(), "readExchangeDatabase", e);

                } else{
                    throw new SyncSourceException("Error reading items", e);
                }
            }

            if (calendar != null) {
                calendar.setModificationStatus(SyncItemState.NEW) ;
                syncItems.add(getSyncItem(calendar)) ;
            }

        }

        return (SyncItem[]) syncItems.toArray(new SyncItem[syncItems.size()]);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getNewSyncItemKeys(Principal principal,
                                            Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getNewSyncItems(Principal principal,
                                      Timestamp since    )
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getNewItemIds();

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "      +
                       sourceURI                   +
                       " - getting new sync items" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.NEW);

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Principal principal,
                                                Timestamp since    )
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getDeletedSyncItems(Principal principal,
                                          Timestamp since    )
    throws SyncSourceException {

        String  [] ids       = null ;
        SyncItem[] syncItems = null ;

        ids = this.getDeleteItemIds();

        int l = ids.length;

        syncItems = new SyncItem[l];

        for (int i = 0; i < l; i++) {

            syncItems[i] = new SyncItemImpl(this, ids[i], SyncItemState.DELETED);
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "          +
                       sourceURI                       +
                       " - getting deleted sync items" );
        }

        return syncItems;

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getUpdatedSyncItems(Principal principal, Timestamp since)
    throws SyncSourceException {

        String ids[] = null;

        ids = this.getUpdateItemIds();

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "          +
                       sourceURI                       +
                       " - getting updated sync items" );
        }

        if(ids == null || !(ids.length > 0)) {
            return new SyncItem[0];
        }

        return filterSyncItems(principal, ids, SyncItemState.UPDATED);

    }

    public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,
                                                Timestamp since)
    throws SyncSourceException {
        return null;
    }

    /**
     * @see SyncSource
     */
    public void removeSyncItem(Principal principal, SyncItem syncItem)
    throws SyncSourceException {

        Calendar calendar    = null  ;

        String   username    = null  ;
        String   credentials = null  ;

        String   id          = null  ;
        String   href        = null  ;

        id          = syncItem.getKey().getKeyAsString()                    ;

        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        href = getHref(id);

        try {
            if (href != null) {
                calendar = new Calendar(id)            ;
                calendar.setHref(href)                 ;
                this.cm.removeCalendar(calendar    ,
                                       username    ,
                                       credentials ,
                                       exchangeFolder)      ;
                toChangedItems(calendar.getId(), ITEM_REMOVE);
            } else {
                toChangedItems(calendar.getId(), ITEM_REMOVE_ERROR);
            }
        } catch (DataAccessException e) {
           toChangedItems(calendar.getId(), ITEM_REMOVE_ERROR);
           throw new SyncSourceException("Error reading items", e);
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "   +
                       sourceURI                +
                       " - removing sync items" );
        }


    }

    /**
     * @see SyncSource
     */
    public void removeSyncItems(Principal principal, SyncItem[] syncItems)
    throws SyncSourceException {

        ArrayList syncItemsInError = new ArrayList();

        for (int i=0, l = syncItems.length ; i < l; ++i) {
            try {
                removeSyncItem(principal, syncItems[i]);
            } catch (SyncSourceException e) {
                syncItemsInError.add(syncItems[i]);
            }
        }  // next i

        if (syncItemsInError.size() > 0) {
            throw new SyncSourceException( "Error deleting the following items: "
                                         + syncItemsInError.toString()
                                         );
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe("Exchange SyncSource "   +
                       sourceURI                +
                       " - removing sync items" );
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem[] getAllSyncItems(Principal principal)
    throws SyncSourceException {
        return filterSyncItems(principal, SyncItemState.NEW);
    }

    /**
     * The same as <i>getSyncItemFromId()</i>
     * @see SyncSource
     */
    public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem)
            throws SyncSourceException {

        String username = ((Sync4jPrincipal) principal).getUsername();
        String credentials = ((Sync4jPrincipal) principal).getEncodedCredentials();
        SyncItem syncTwin = null;

        try {
            this.cm = getCalendarManager();
            Calendar calendar = getCalendarFromSyncItem(syncItem);
            Calendar twin = this.cm.getCalendarTwin(calendar,
                    username,
                    credentials,
                    exchangeFolder);

            if (twin != null) {
                syncTwin = getSyncItem(twin);
            }

        } catch (DataAccessException ex) {
            throw new SyncSourceException("Error getting twin", ex);
        }

        return syncTwin;
    }


    /**
     * The same as <i>getSyncItemFromIds()</i>
     *
     * @see SyncSource
     */
    public SyncItem[] getSyncItemsFromTwins(Principal principal,
                                            SyncItem[] syncItems) throws
            SyncSourceException {

        ArrayList items = new ArrayList();

        for (int i = 0, l = syncItems.length; i < l; ++i) {
            items.add(getSyncItemFromTwin(principal, syncItems[i]));
        }

        return (SyncItem[]) items.toArray(new SyncItem[items.size()]);

    }
    // --------------------------------------------------------- Private methods

    private Calendar getCalendar(String    id      ,
                                 String    content ,
                                 Date      t       )
    throws SyncSourceException {

        Calendar c = new Calendar(id);

        c.setLastUpdate(t) ;

        if (content != null && content.length() > 0) {

            if (isEncode()) {
                content = new String (Base64.decode(content));
            }

            if (TYPE_ICAL.equals(getType())) {
                ical2Calendar(content, c);
            } else {
                sife2Calendar(content, c);
            }
        }

        return c;

    }



    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, ids, state
     *
     * @param principal principal
     * @param ids     an array containing calendar id
     * @param state   calendar state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       String[]   ids       ,
                                       char       state)
    throws SyncSourceException {

       Calendar[] calendars   = null ;

       String     username    = null ;
       String     credentials = null ;

       username    = ((Sync4jPrincipal) principal).getUsername()           ;
       credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

       try {

           this.cm = getCalendarManager();


           calendars = this.cm.getCalendars (username    ,
                                             credentials ,
                                             ids         ,
                                             this.exchangeFolder) ;

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading calendar: " +
                                          e.getMessage(), e);
        }

        return getSyncItems(calendars, state);

    }

    /**
     * Filters the SyncItems in the synchronization database (after a refresh)
     * based on the given principal, state
     *
     * @param principal principal
     * @param ids     an array containing calendar id
     * @param state   calendar state
     *
     * @throws SyncSourceException in case of IO errors
     *
     * @return an array of SyncItem objects whose state is equal to the given
     *         state.
     */
    private SyncItem[] filterSyncItems(Principal  principal ,
                                       char       state)
    throws SyncSourceException {

       Calendar[] calendars    = null ;

       String     username     = null ;
       String     credentials  = null ;

       username    = ((Sync4jPrincipal) principal).getUsername()           ;
       credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

       try {

           this.cm = getCalendarManager();

           calendars = this.cm.getAllCalendars (username    ,
                                                credentials ,
                                                this.exchangeFolder) ;

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error reading calendar: " +
                                           e.getMessage(), e);
        }

        return getSyncItems(calendars, state);

    }

    /**
     * Create SyncItem array from Calendar array
     *
     * @param calendars
     * @param state
     * @return SyncItem array
     * @trows SyncSourceException
     */
    private SyncItem[] getSyncItems(Calendar[] calendars,
                                    char       state)
    throws SyncSourceException {

        SyncItem[] syncItems = null;

        int l = calendars.length;

        syncItems = new SyncItem [l];

        for (int i=0; i < l; i++) {
            calendars [i].setModificationStatus(state)             ;
            syncItems [i] = getSyncItem(calendars[i]) ;
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Calendar
     *
     * @param calendar
     * @return SyncItem
     * @trows SyncSourceException
     */
    private SyncItem getSyncItem(Calendar exchangeCalendar)
    throws SyncSourceException {

        SyncItem syncItem  = null ;
        String   content   = null ;

        sync4j.foundation.pdi.event.Calendar
                    calendar = new sync4j.foundation.pdi.event.Calendar();

        syncItem = new SyncItemImpl(this                                      ,
                                    exchangeCalendar.getId                 () ,
                                    exchangeCalendar.getModificationStatus ());

        calendar.setEvent(exchangeCalendar);
        if (TYPE_ICAL.equals(getType())) {
            content = calendar2ical(calendar);
        } else {
            content = calendar2sife(calendar);
        }

        if (this.isEncode()) {
            syncItem.setProperty(
                new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                 Base64.encode(content.getBytes())
            ));

           syncItem.setProperty(
               new SyncProperty(SyncItem.PROPERTY_FORMAT,"b64")
           );

        } else {
            syncItem.setProperty(
                new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT,
                                 content.getBytes())
            );
        }

        return syncItem;

    }

    /**
     * Converts the given calendar into a ical String
     *
     * @param c the calendar to convert
     *
     * @return the ical
     *
     * @throws SyncSourceException in case of convertion errors
     */
    private String calendar2ical(sync4j.foundation.pdi.event.Calendar c)
    throws SyncSourceException {
        try {
            return new CalendarToIcalendar(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException( "Conversion error for item "
                                          + ((Calendar)c.getEvent()).getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        }
    }

    /**
     * Converts the given calendar into a sife String
     *
     * @param c the calendar to convert
     *
     * @return the sife document
     *
     * @throws SyncSourceException in case of convertion errors
     */
    private String calendar2sife(sync4j.foundation.pdi.event.Calendar c)
    throws SyncSourceException {
        try {
            return new CalendarToXML(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException( "Convertion error for item "
                                          + ((Calendar)c.getEvent()).getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        }
    }


    /**
     *
     * Fill the given CalEvent object with the properties in the given ical
     *
     * @param content iCal content
     * @param c the calendar to feed
     *
     * @throws SyncSourceException
     */
    private void ical2Calendar (String content, Calendar c)
    throws SyncSourceException {

        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length()==0)) {
            return;
        }


        ByteArrayInputStream is     = null ;
        ICalendarParser      parser = null ;

        sync4j.foundation.pdi.event.Calendar calendar = null ;
        sync4j.foundation.pdi.event.Event    event    = null ;

        try {

            /**
             * This is used to replace the String "=\r\n" with the
             * String "\r\n ": This replace is necessary because some devices do
             * not send the correct line delimiter for long lines.
             */
            content = content.replaceAll("=\r\n","\r\n ");

            is = new ByteArrayInputStream (content.getBytes());

            //converting the vcard into a Contact object

            parser = new ICalendarParser(is, deviceTimeZoneDescr, deviceCharset);
            parser.setLogger(log);

            calendar = (sync4j.foundation.pdi.event.Calendar)parser.iCalendar();

            //
            // We now have to convert a portal calendar into a foundation calendar
            //
            event = calendar.getEvent();

            c.setReminder(event.getReminder());
            c.setCategories(event.getCategories());
            c.setClassEvent(event.getClassEvent());
            c.setDescription(event.getDescription());
            c.setLatitude(event.getLatitude());
            c.setLongitude(event.getLongitude());
            c.setLocation(event.getLocation());
            c.setPriority(event.getPriority());
            c.setStatus(event.getStatus());
            c.setSummary(event.getSummary());
            c.setDtEnd(event.getDtEnd());
            c.setDtStart(event.getDtStart());
            c.setDuration(event.getDuration());
            c.setTransp(event.getTransp());
            c.setOrganizer(event.getOrganizer());
            c.setUrl(event.getUrl());
            c.setUid(event.getUid());
            c.setRrule(event.getRrule());
            c.setContact(event.getContact());
            c.setCreated(event.getCreated());
            c.setDtStamp(event.getDtStamp());
            c.setLastModified(event.getLastModified());
            c.setSequence(event.getSequence());
            c.setXTag(event.getXTags());
            c.setAllDay(new Boolean(event.isAllDay()));
            c.setMeetingStatus(event.getMeetingStatus());
            c.setMileage(event.getMileage());
            c.setReminder(event.getReminder());
            c.setReplyTime(event.getReplyTime());
            c.setRecurrencePattern(event.getRecurrencePattern());

        } catch (Exception e) {
            throw new SyncSourceException( "Convertion error for item "
                                          + c.getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        } finally {
            if (is != null) try { is.close(); } catch (Exception e) {}
        }
    }


    /**
     *
     * Fill the given CalEvent object with the properties in the given sife
     * document
     *
     * @param content iCal content
     * @param c the calendar to feed
     *
     * @throws SyncSourceException
     */
    private void sife2Calendar (String content, Calendar c)
    throws SyncSourceException {
        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length()==0)) {
            return;
        }


        ByteArrayInputStream  is         = null ;
        XMLEventParser        parser     = null ;

        sync4j.foundation.pdi.event.Calendar calendar = null ;
        sync4j.foundation.pdi.event.Event    event    = null ;

        try {
            is = new ByteArrayInputStream (content.getBytes());

            //converting the vcard into a Contact object

            parser = new XMLEventParser(is);

            calendar = (sync4j.foundation.pdi.event.Calendar)parser.parse();

            //
            // We now have to convert a portal contact into a foundation contact
            //
            event = calendar.getEvent();

            c.setReminder(event.getReminder());
            c.setCategories(event.getCategories());
            c.setClassEvent(event.getClassEvent());
            c.setDescription(event.getDescription());
            c.setLatitude(event.getLatitude());
            c.setLongitude(event.getLongitude());
            c.setLocation(event.getLocation());
            c.setPriority(event.getPriority());
            c.setStatus(event.getStatus());
            c.setSummary(event.getSummary());
            c.setDtEnd(event.getDtEnd());
            c.setDtStart(event.getDtStart());
            c.setDuration(event.getDuration());
            c.setTransp(event.getTransp());
            c.setOrganizer(event.getOrganizer());
            c.setUrl(event.getUrl());
            c.setUid(event.getUid());
            c.setRrule(event.getRrule());
            c.setContact(event.getContact());
            c.setCreated(event.getCreated());
            c.setDtStamp(event.getDtStamp());
            c.setLastModified(event.getLastModified());
            c.setSequence(event.getSequence());
            c.setXTag(event.getXTags());
            c.setAllDay(new Boolean(event.isAllDay()));
            c.setMeetingStatus(event.getMeetingStatus());
            c.setMileage(event.getMileage());
            c.setReminder(event.getReminder());
            c.setReplyTime(event.getReplyTime());
            c.setRecurrencePattern(event.getRecurrencePattern());

        } catch (Exception e) {
            throw new SyncSourceException( "Parsing error for item "
                                          + c.getId()
                                          + ": "
                                          + e.getMessage()
                                          , e
                                          );
        } finally {
            if (is != null) try { is.close(); } catch (Exception e) {}
        }
    }

    /**
     *
     * Create CalendarManager istance
     * (if not already exist)
     *
     * @trows DataAccessException
     */
    private CalendarManager getCalendarManager()
    throws DataAccessException {

        if (this.cm == null) {
            this.cm = new CalendarManager(this.getHost() ,
                                          this.getPort() ,
                                          this.deviceTimeZone);
        }

        return this.cm;

    }

    /**
     * Creates a new Calendar from the given syncItem
     * @param syncItem SyncItem
     * @return Calendar
     */
    private Calendar getCalendarFromSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        String content    = null;
        Calendar calendar = null;

        String itemKey = syncItem.getKey().getKeyAsString();

        byte[] itemContent =
                (byte[]) syncItem.getPropertyValue(
                        SyncItem.PROPERTY_BINARY_CONTENT);

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        content = new String(itemContent);

        Date t =
                (Timestamp) syncItem.getPropertyValue(SyncItem.PROPERTY_TIMESTAMP);

        calendar = getCalendar(itemKey, content, t);

        return calendar;
    }
}
